/// This file is part of the X3DToolKit library
/// Copyright (C) 2002-2004  Yannick Le Goc (legoc@imag.fr)

/// http://artis.imag.fr/Members/Yannick.Legoc/X3D/

/// This library is free software; you can redistribute it and/or
/// modify it under the terms of the GNU Lesser General Public
/// License as published by the Free Software Foundation; either
/// version 2.1 of the License, or (at your option) any later version.

/// This library is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/// Lesser General Public License for more details.

/// You should have received a copy of the GNU Lesser General Public
/// License along with this library; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
///                            Route.h                                       ///
////////////////////////////////////////////////////////////////////////////////

#ifndef ROUTE_H
#define ROUTE_H

#include "Platform.h"

#include <map>
#include <list>
#include <iostream>

namespace RC {

template<class A, class B, class T> class TRoute;
template<class A, class T> class RouteSender;

/// Class defining a signal/slot mechanism based on template functions.

////////////////////////////////////////////////////////////////////////////////
/// Route

class Route
{
public:
  virtual void transmit() = 0;

#ifdef TEMPLATE_SPECIALIZATION_SUPPORTED
  template<class A, class SA, class B, class SB, class T>
  inline static void connect(A *a, T (SA:: *getPtr)() const, B *b, void (SB:: *setPtr)(T t))
  {
    //std::cout << "connecting..." << std::endl;
    RouteSender<SA, T>::connect(a, getPtr, static_cast<SB *>(b), setPtr);  
  };

  template<class A, class SA, class B, class SB>
  inline static void connect(A *a, void (SA:: *getPtr)() const, B *b, void (SB:: *setPtr)())
  {
    //std::cout << "connecting..." << std::endl;
    RouteSender<SA, void>::connect(a, getPtr, static_cast<SB *>(b), setPtr);  
  };

  template<class A, class SA, class B, class SB, class T>
  inline static void disconnect(A *a, T (SA:: *getPtr)() const, B *b, void (SB:: *setPtr)(T t))
  {
    //std::cout << "disconnecting...";
    RouteSender<SA, T>::disconnect(a, getPtr, static_cast<SB *>(b), setPtr);  
  };

  template<class A, class SA, class B, class SB>
  inline static void disconnect(A *a, void (SA:: *getPtr)() const, B *b, void (SB:: *setPtr)())
  {
    //std::cout << "disconnecting...";
    RouteSender<SA, void>::disconnect(a, getPtr, static_cast<SB *>(b), setPtr);  
  };

  template<class A, class SA, class T>
  inline static void send(A *a, T (SA:: *getPtr)() const)
  {
    RouteSender<SA, T>::send(a, getPtr);  
  };
#endif  
};

#ifdef TEMPLATE_SPECIALIZATION_SUPPORTED

////////////////////////////////////////////////////////////////////////////////
/// RouteSender

template<class A, class T>
class RouteSender
{
private:
  typedef std::list<std::pair<void (A:: *)() const, Route *> > RouteMap;
  typedef std::map<A *, RouteMap> RouteMapMap;

public:
  template<class B>
  static void connect(A *a, T (A:: *getPtr)() const, B *b, void (B:: *setPtr)(T t));

  template<class B>
  static void disconnect(A *a, T (A:: *getPtr)() const, B *b, void (B:: *setPtr)(T t));

  static void send(A *a, T (A:: *getPtr)() const);
  
private:
  static RouteMapMap _routeMap;
};

template<class A>
class RouteSender<A, void>
{
private:
  typedef std::list<std::pair<void (A:: *)() const, Route *> > RouteMap;
  typedef std::map<A *, RouteMap> RouteMapMap;

public:
  template<class B>
  static void connect(A *a, void (A:: *getPtr)() const, B *b, void (B:: *setPtr)());

  template<class B>
  static void disconnect(A *a, void (A:: *getPtr)() const, B *b, void (B:: *setPtr)());

  static void send(A *a, void (A:: *getPtr)() const);
  
private:
  static RouteMapMap _routeMap;
};

////////////////////////////////////////////////////////////////////////////////
/// TRoute

template<class A, class B, class T>
class TRoute : public Route
{
  template<class X, class Y> friend class RouteSender;
public:
  TRoute(A *a, T (A:: *getPtr)() const, B *b, void (B:: *setPtr)(T t));

  void transmit();

private:
  A *_a;
  T (A:: *_getPtr)() const;
  B *_b;
  void (B:: *_setPtr)(T t);
};

template<class A, class B>
class TRoute<A, B, void> : public Route
{
  template<class X, class Y> friend class RouteSender;
public:
  TRoute(A *a, void (A:: *getPtr)() const, B *b, void (B:: *setPtr)());

  void transmit();

private:
  A *_a;
  void (A:: *_getPtr)() const;
  B *_b;
  void (B:: *_setPtr)();
};


template<class A, class T>  
template<class B>
void RouteSender<A, T>::connect(A *a, T (A:: *getPtr)() const, B *b, void (B:: *setPtr)(T t))
{
  Route *r = new TRoute<A, B, T>(a, getPtr, b, setPtr);
  _routeMap[a].push_back(std::pair<void (A:: *)() const, Route *>(reinterpret_cast<void (A:: *)() const>(getPtr), r));
}

template<class A, class T>
template<class B>
void RouteSender<A, T>::disconnect(A *a, T (A:: *getPtr)() const, B *b, void (B:: *setPtr)(T t))
{
  for (typename RouteMap::iterator it = _routeMap[a].begin(); it != _routeMap[a].end(); ++it)
  {
    if ((*it).first == reinterpret_cast<void (A:: *)() const>(getPtr))
    {
      TRoute<A, B, T> *tr = dynamic_cast<TRoute<A, B, T> *>((*it).second);
    
      if ((tr != 0) 
       && (b == tr->_b)
       && (setPtr == tr->_setPtr))
      {
        _routeMap[a].erase(it);
        delete tr;
        if (_routeMap[a].empty())  
          _routeMap.erase(a);
        return;           
      }
    }
  }  
}

template<class A, class T>
void RouteSender<A, T>::send(A *a, T (A:: *getPtr)() const)
{
  //std::cout << "sending..." << std::endl;      
  for (typename RouteMap::iterator it = _routeMap[a].begin(); it != _routeMap[a].end(); ++it)
  {
    if ((*it).first == reinterpret_cast<void (A:: *)() const>(getPtr))
      (*it).second->transmit();
  }  
}

template<class A>
template<class B>
void RouteSender<A, void>::connect(A *a, void (A:: *getPtr)() const, B *b, void (B:: *setPtr)())
{
  Route *r = new TRoute<A, B, void>(a, getPtr, b, setPtr);
  _routeMap[a].push_back(std::pair<void (A:: *)() const, Route *>(reinterpret_cast<void (A:: *)() const>(getPtr), r));
}

template<class A>
template<class B>
void RouteSender<A, void>::disconnect(A *a, void (A:: *getPtr)() const, B *b, void (B:: *setPtr)())
{
  for (typename RouteMap::iterator it = _routeMap[a].begin(); it != _routeMap[a].end(); ++it)
  {
    if ((*it).first == reinterpret_cast<void (A:: *)() const>(getPtr))
    {
      TRoute<A, B, void> *tr = dynamic_cast<TRoute<A, B, void> *>((*it).second);
    
      if ((tr != 0) 
       && (b == tr->_b)
       && (setPtr == tr->_setPtr))
      {
        _routeMap[a].erase(it);
        delete tr;
        if (_routeMap[a].empty())  
          _routeMap.erase(a);
        return;
      }
    }
  }
}

template<class A>
void RouteSender<A, void>::send(A *a, void (A:: *getPtr)() const)
{
  //std::cout << "sending void..." << std::endl;    
  for (typename RouteMap::iterator it = _routeMap[a].begin(); it != _routeMap[a].end(); ++it)
  {
    if ((*it).first == reinterpret_cast<void (A:: *)() const>(getPtr))
      (*it).second->transmit();
  } 
}

template<class A, class T>
typename RouteSender<A, T>::RouteMapMap RouteSender<A, T>::_routeMap;

template<class A>
typename RouteSender<A, void>::RouteMapMap RouteSender<A, void>::_routeMap;

////////////////////////////////////////////////////////////////////////////////
/// TRoute

template<class A, class B, class T>
TRoute<A, B, T>::TRoute(A *a, T (A:: *getPtr)() const, B *b, void (B:: *setPtr)(T t))
: _a(a), _getPtr(getPtr), _b(b), _setPtr(setPtr)
{
}

template<class A, class B, class T>
void TRoute<A, B, T>::transmit()
{
  //std::cout << "transmitting..." << std::endl;
  (_b->*_setPtr)((_a->*_getPtr)());
}

template<class A, class B>
TRoute<A, B, void>::TRoute(A *a, void (A:: *getPtr)() const, B *b, void (B:: *setPtr)())
: _a(a), _getPtr(getPtr), _b(b), _setPtr(setPtr)
{
}

template<class A, class B>
void TRoute<A, B, void>::transmit()
{
  //std::cout << "transmitting..." << std::endl;
  (_a->*_getPtr)();
  (_b->*_setPtr)();
}

#endif

}

#endif

